#include "math.h"
#include "mex.h"
#include "surflib.h"
#include <iostream>
#define VAR(x)  #x " = " << x

using namespace std;
using namespace surf;
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    // parameters:
    // - image
    // - points
    // - upright  (true to switch off rotationInvariance)
    // - extended (true for SURF-128)
    // - size (descriptor size)
    
    // --- get the stuff from matlab
    bool doubleImageSize=false, upright=false, extended=false, verbose=false;
    bool bLaplacian=true;
    
    double thres=4.0;
    int initLobe=3;
    int samplingStep=2;
    int octaves=4;
    
    int indexSize=4;
    const mxArray *mxImage=prhs[0], *mxPoints=prhs[1];
    
    int W=mxGetN(mxImage);
    int H=mxGetM(mxImage);
    double *mxPixels=mxGetPr(mxImage);
    
    // -- parsing of optional parameters
    if(nrhs>2) thres=mxGetScalar(prhs[2]);
    if(nrhs>3) doubleImageSize=mxGetLogicals(prhs[3])[0];
    if(nrhs>4) initLobe=(int)mxGetScalar(prhs[4]);
    if(nrhs>5) samplingStep=(int)mxGetScalar(prhs[5]);
    if(nrhs>6) octaves=(int)mxGetScalar(prhs[6]);
    if(nrhs>7) upright=mxGetLogicals(prhs[7])[0];
    if(nrhs>8) extended=mxGetLogicals(prhs[8])[0];
    if(nrhs>9) indexSize=(int)mxGetScalar(prhs[9]);
    if(nrhs>10) bLaplacian=mxGetLogicals(prhs[10])[0];
    if(nrhs>11) verbose=mxGetLogicals(prhs[11])[0];
    
    
    if(verbose) {
        cout << VAR(thres) << " " << VAR(doubleImageSize) << " " << VAR(initLobe) << " "
                << VAR(samplingStep) << " " << VAR(octaves) << " " << VAR(upright) << " "
                << VAR(extended) << " " << VAR(indexSize) << " " << VAR(bLaplacian) << " "
                VAR(verbose) << endl;
    }
    
    // --- put image into surf structure
    Image img(W, H);
    for(uint x=0;x<W;++x) {
        for(uint y=0;y<H;++y) {
            img.setPix(x, y, mxPixels[x*H+y]);
        }
    }
    
    
    // get integral image
    Image iimage(&img, doubleImageSize);
    
    // --- put points into surf structure
    int N=mxGetM(mxPoints);
    vector<Ipoint> ipts(N);
    if(mxGetN(mxPoints)!=5) {
        cerr << "interest point array must be five columns [x y a b c]" << endl;
        return;
    }
    
    double *mxpts=mxGetPr(mxPoints);
    double x, y, a, b, c;
    for(int n=0;n<N;++n) {
        x=mxpts[n+0*N];
        y=mxpts[n+1*N];
        a=mxpts[n+2*N];
        b=mxpts[n+3*N];
        c=mxpts[n+4*N];
        
        if (x>W) {cerr << "x=" << x << " > W=" << W << endl;}
        if (y>H) {cerr << "y=" << y << " > H=" << H << endl;}
        
        double det = sqrt((a-c)*(a-c) + 4.0*b*b);
        double e1 = 0.5*(a+c + det);
        double e2 = 0.5*(a+c - det);
        double l1 = (1.0/sqrt(e1));
        double l2 = (1.0/sqrt(e2));
        double sc = sqrt( l1*l2 );
        
        ipts[n].x     = x;
        ipts[n].y     = y;
        ipts[n].scale = sc/2.5;
        ipts[n].laplace=0;
    }
    
    if(verbose){ cout << "Extracting SURF features at " << VAR(N) << " interestpoints from " << VAR(W) << " x " << VAR(H) << " image." << endl; }
    
    // Initialise the SURF descriptor
    Surf des(&iimage, /* pointer to integral image */
            doubleImageSize, /* double image size flag */
            upright, /* rotation invariance or upright */
            extended, /* use the extended descriptor */
            indexSize /* square size of the descriptor window (default 4x4)*/);
    
    
    int VLength = des.getVectLength();
    
    for (unsigned n=0; n<ipts.size(); n++){
        des.setIpoint(&ipts[n]);
        
        des.assignOrientation();
        des.makeDescriptor();
    }
    
    // -- game over
    // copy stuff into matlab data structures
    
    if(bLaplacian){
        plhs[0]=mxCreateDoubleMatrix( ipts.size(), VLength+1, mxREAL);
    } else {
        plhs[0]=mxCreateDoubleMatrix( ipts.size(), VLength, mxREAL);
    }
    
    double *output=mxGetPr(plhs[0]);
    
    for (unsigned n=0; n<ipts.size(); n++){
        if(bLaplacian) {
            output[n]=ipts[n].laplace;
            for(unsigned d=0; d<VLength;++d) {
                output[N*(d+1)+n]=ipts[n].ivec[d];
            }
        } else {
            for(unsigned d=0; d<VLength;++d) {
                output[N*d+n]=ipts[n].ivec[d];
            }
        }
    }
}

