function obj = classcutSegmentImagesDir(soft_dir,d,image_type, params)

% segments all images in directory d,
% saves .ppm file with segmentations, and one .mat file with pixel coordinates
% within each segment for all images.
% 
% uses sof_dir to invoke 'segment' exec 
%
% uses Felzenswalb and Huttelocher image-segmentation c++ excutable.
% image_type can be anything, as this function automatically converts it to
% .ppm (only format supported by the exec).
%
% params = [sigma k min]
% as in the exec's readme, but tuned for images of about 300x200 pixels;
% sigma and min are automatically scaled to the image size
% (best params in gen seem to be (for a 300x200 img): sigma = 0.5, k = 500, min = 200 (as given by Felzenswalb),
% but then prop to sqrt(image area))


% save current working directory and LD_LIBRARY_PATH (avoid conflicts with
% what 'segment' is dyn linked to)
org_dir = pwd;
org_ldlp = getenv('LD_LIBRARY_PATH');
setenv('LD_LIBRARY_PATH', '');

% prepare params
basis_sigma = params(1);
basis_k = params(2);
basis_min_area = params(3);

% get list of images
cd(d);
images = dir(['*.' image_type]);

disp(['Images found: ']);
dir(['*.' image_type]);               % must display this way, because variable is a struct

% segment images
nb_images = 0;
clear obj;

for i = images'
  nb_images = nb_images+1;
  disp(i.name);
  disp(['--------------------']);
              
  % shortcuts
  name = i.name;    
  i_base = name(1:(length(name)-length(image_type)-1));
  
  segm_fname = [i_base '_segm.ppm'];    
  pb_fname = [i_base '.pb'];  
  
  if not(exist(segm_fname,'file'))

      % convert image to ppm
      if not(strcmp(image_type, 'ppm'))
        cmd = [ 'convert "' i.name '" "' i_base '.ppm"' ];
        disp(['Invoking command:']); disp(cmd); system(cmd);
      end

      % setting params
      I = imread([i_base '.ppm']);
      Iarea = size(I,1)*size(I,2);
      sf = sqrt(Iarea/(300*200));
      sigma = basis_sigma*sf;
      min_area = basis_min_area*sf;
      k = basis_k;                             % not scaled

      % segment image
      disp('Segmenting image ....');      
      cmd = [soft_dir '/objectness-release/segment/segment ' num2str(sigma) ' ' num2str(k) ' ' num2str(min_area) ' "' i_base '.ppm' '" "' segm_fname '"' ];
      disp(['Invoking command:']); disp(cmd); system(cmd);

      % delete image ppm
      if not(strcmp(image_type, 'ppm'))
        delete([i_base '.ppm']);
      end
  else    % segmentation file found
      disp([segm_fname ' found -> not recomputing segmentation']);
  end
  
  if not(exist(pb_fname,'file'))
      %generate BSE .pb files
      if not(strcmp(image_type, 'jpeg'))
        cmd = [ 'convert "' i.name '" "' i_base '.jpeg"' ];
        disp('Invoking command:'); disp(cmd); system(cmd);     
      end
            
      cmd = [soft_dir '/BSE-1.2/pixelfeatures -image ' d '/' i_base '.jpeg -pbfile ' d '/' i_base '.pb'];
      cmd = strrep(cmd,'//','/');
      disp('Invoking command:'); disp(cmd); system(cmd);
      
      %delete image jpeg
      if not(strcmp(image_type, 'jpeg'))
        delete([i_base '.jpeg']);
      end
  else
      disp([pb_fname ' found -> not recomputing segmentation']);
  end
   
  
  % get pixel coordinates within each segment  
  disp('Numerizing segment labels and preparing segms struct ...');  
  I = imread([i_base '.' image_type]);  
  S = imread(segm_fname);
  N = classcutNumerizeLabels(S);
  obj(nb_images).segms = classcutSegmentCoordsObjectnessNeighbors(N,I);    
  obj(nb_images).pb = read_array([d '/' images(nb_images).name(1:end-length(image_type)) 'pb']);
  cmd = [d '/' images(nb_images).name(1:end-length(image_type)) 'pb'];
  cmd = strrep(cmd,'//','/');
  obj(nb_images).pb = read_array(cmd);
  h = fspecial('gaussian',[4 4],0.5);
  obj(nb_images).pbblurred = imfilter(obj(nb_images).pb,h,'replicate');
    
  obj(nb_images).name = [d '/' i.name];   
  obj(nb_images).segm_labels = N;  
end

obj = AdjacentSegments(obj, 'adjacent');
%obj = AdjacentSegments(obj, 'near_adjacent');
%obj = AdjacentSegments(obj, 'heavily_adjacent');
obj = classcutComputeGradientBorder(obj,'adjacent');
mat_fname = 'obj.mat';
disp(['Saving segmentation to ' mat_fname]);
save(mat_fname, 'obj','-v7');

% restore current working directory and LD_LIBRARY_PATH
cd(org_dir);
setenv('LD_LIBRARY_PATH', org_ldlp);
